import Branch from "./branch";
import { metroTemplate, blackArrowTemplate, } from "./template";
import Refs from "./refs";
import { booleanOptionOr, numberOptionOr, pick } from "./utils";
export var OrientationsEnum;
(function (OrientationsEnum) {
    OrientationsEnum["VerticalReverse"] = "vertical-reverse";
    OrientationsEnum["Horizontal"] = "horizontal";
    OrientationsEnum["HorizontalReverse"] = "horizontal-reverse";
})(OrientationsEnum || (OrientationsEnum = {}));
export var ModeEnum;
(function (ModeEnum) {
    ModeEnum["Compact"] = "compact";
})(ModeEnum || (ModeEnum = {}));
export var TemplateEnum;
(function (TemplateEnum) {
    TemplateEnum["Metro"] = "metro";
    TemplateEnum["BlackArrow"] = "blackarrow";
})(TemplateEnum || (TemplateEnum = {}));
export class GitgraphCore {
    constructor(options = {}) {
        this.refs = new Refs();
        this.tags = new Refs();
        this.commits = [];
        this.branches = new Map();
        this.columns = [];
        this.rows = new Map();
        this.maxRow = 0;
        this.listeners = [];
        if (typeof options.template === "string") {
            this.template = {
                [TemplateEnum.BlackArrow]: blackArrowTemplate,
                [TemplateEnum.Metro]: metroTemplate,
            }[options.template];
        }
        else if (options.template) {
            this.template = options.template;
        }
        else {
            this.template = metroTemplate;
        }
        this.currentBranch = this.branch("master");
        this.orientation = options.orientation;
        this.isVertical = [
            undefined,
            OrientationsEnum.VerticalReverse,
        ].includes(this.orientation);
        this.reverseArrow = booleanOptionOr(options.reverseArrow, false);
        this.initCommitOffsetX = numberOptionOr(options.initCommitOffsetX, 0);
        this.initCommitOffsetY = numberOptionOr(options.initCommitOffsetY, 0);
        this.mode = options.mode;
        this.author = options.author || "Sergio Flores <saxo-guy@epic.com>";
        this.commitMessage =
            options.commitMessage || "He doesn't like George Michael! Boooo!";
        this.withRefsAndTags = this.withRefsAndTags.bind(this);
        this.withPosition = this.withPosition.bind(this);
        this.withColor = this.withColor.bind(this);
        this.withBranches = this.withBranches.bind(this);
        this.calculateRows = this.calculateRows.bind(this);
        this.initBranchesPaths = this.initBranchesPaths.bind(this);
        this.addMergeCommitsIntoBranchesPaths = this.addMergeCommitsIntoBranchesPaths.bind(this);
    }
    getRenderedData() {
        let commits = this.commits.map(this.withRefsAndTags);
        commits = this.withBranches(commits, { firstParentOnly: true });
        this.calculateRows(commits);
        commits = commits.map(this.withPosition).map(this.withColor);
        const flatBranchesPaths = commits.reduce(this.initBranchesPaths, new Map());
        commits = this.withBranches(commits);
        this.addMergeCommitsIntoBranchesPaths(commits, flatBranchesPaths);
        const branchesPaths = this.smoothBranchesPaths(flatBranchesPaths);
        Array.from(branchesPaths).forEach(([branch], i) => {
            const defaultColor = this.template.colors[i % this.template.colors.length];
            branch.computedColor = branch.style.color || defaultColor;
        });
        const numberOfColumns = Array.from(branchesPaths).length;
        const commitMessagesX = numberOfColumns * this.template.branch.spacing;
        return { commits, branchesPaths, commitMessagesX };
    }
    commit(options) {
        this.currentBranch.commit(options);
        return this;
    }
    branch(args) {
        const parentCommitHash = this.refs.get("HEAD");
        let options = {
            gitgraph: this,
            name: "",
            parentCommitHash,
            style: this.template.branch,
        };
        if (typeof args === "string") {
            options.name = args;
        }
        else {
            options = Object.assign({}, options, args);
        }
        const branch = new Branch(options);
        this.branches.set(branch.name, branch);
        return branch;
    }
    clear() {
        this.refs = new Refs();
        this.tags = new Refs();
        this.commits = [];
        this.columns = [];
        this.branches = new Map();
        this.currentBranch = this.branch("master");
        this.next();
        return this;
    }
    tag(name, ref = this.refs.get("HEAD")) {
        if (typeof ref === "string") {
            const commitHashOrRefs = this.refs.get(ref);
            if (!commitHashOrRefs) {
                throw new Error("this ref not exists");
            }
            else if (typeof commitHashOrRefs === "object") {
                this.tags.set(name, ref);
            }
            else {
                this.tags.set(name, commitHashOrRefs);
            }
        }
        else {
            this.tags.set(name, ref.hash);
        }
        this.next();
        return this;
    }
    subscribe(listener) {
        this.listeners.push(listener);
        let isSubscribed = true;
        return () => {
            if (!isSubscribed)
                return;
            isSubscribed = false;
            const index = this.listeners.indexOf(listener);
            this.listeners.splice(index, 1);
        };
    }
    next() {
        this.listeners.forEach(listener => listener());
    }
    withRefsAndTags(commit) {
        return Object.assign({}, commit, { refs: this.refs.get(commit.hash) || [], tags: this.tags.get(commit.hash) || [] });
    }
    withBranches(commits, { firstParentOnly } = { firstParentOnly: false }) {
        const branches = Array.from(this.refs)
            .filter(([key, value]) => typeof value === "string" && key !== "HEAD")
            .map(([key]) => key);
        const refs = new Map();
        const queue = [];
        branches.forEach((branch) => {
            queue.push(this.refs.get(branch));
            while (queue.length > 0) {
                const currentHash = queue.pop();
                const current = commits.find(({ hash }) => hash === currentHash);
                const prevBranches = refs.get(currentHash) || new Set();
                prevBranches.add(branch);
                refs.set(currentHash, prevBranches);
                if (current.parents.length > 0) {
                    firstParentOnly
                        ? queue.push(current.parents[0])
                        : queue.push(...current.parents);
                }
            }
        });
        return commits.map(commit => (Object.assign({}, commit, { branches: Array.from((refs.get(commit.hash) || new Set()).values()) })));
    }
    calculateRows(commits) {
        this.rows = new Map();
        this.maxRow = 0;
        commits.forEach((commit, i) => {
            if (this.mode === ModeEnum.Compact) {
                if (i === 0)
                    return this.rows.set(commit.hash, i);
                const parentRow = this.rows.get(commit.parents[0]);
                const historyParent = commits[i - 1];
                let newRow = Math.max(parentRow + 1, this.rows.get(historyParent.hash));
                const isMergeCommit = (commit.parents.length > 1);
                if (isMergeCommit) {
                    const mergeTargetParentRow = this.rows.get(commit.parents[1]);
                    if (parentRow < mergeTargetParentRow)
                        newRow++;
                }
                this.rows.set(commit.hash, newRow);
                this.maxRow = Math.max(this.maxRow, newRow + 1);
            }
            else {
                this.rows.set(commit.hash, i);
                this.maxRow = Math.max(this.maxRow, i + 1);
            }
        });
    }
    withColor(commit) {
        const branch = commit.branches[0];
        const column = this.columns.findIndex(col => col === branch);
        const defaultColor = this.template.colors[column % this.template.colors.length];
        return Object.assign({}, commit, { style: Object.assign({}, commit.style, { color: commit.style.color || defaultColor, tag: Object.assign({}, commit.style.tag, { color: commit.style.tag.color || defaultColor }), dot: Object.assign({}, commit.style.dot, { color: commit.style.dot.color || defaultColor }), message: Object.assign({}, commit.style.message, { color: commit.style.message.color || defaultColor }) }) });
    }
    withPosition(commit) {
        const branch = commit.branches[0];
        if (!this.columns.includes(branch))
            this.columns.push(branch);
        const column = this.columns.findIndex(col => col === branch);
        const row = this.rows.get(commit.hash);
        switch (this.orientation) {
            default:
                return Object.assign({}, commit, { x: this.initCommitOffsetX + this.template.branch.spacing * column, y: this.initCommitOffsetY +
                        this.template.commit.spacing * (this.maxRow - 1 - row) });
            case OrientationsEnum.VerticalReverse:
                return Object.assign({}, commit, { x: this.initCommitOffsetX + this.template.branch.spacing * column, y: this.initCommitOffsetY + this.template.commit.spacing * row });
            case OrientationsEnum.Horizontal:
                return Object.assign({}, commit, { x: this.initCommitOffsetX + this.template.commit.spacing * row, y: this.initCommitOffsetY + this.template.branch.spacing * column });
            case OrientationsEnum.HorizontalReverse:
                return Object.assign({}, commit, { x: this.initCommitOffsetX +
                        this.template.commit.spacing * (this.maxRow - 1 - row), y: this.initCommitOffsetY + this.template.branch.spacing * column });
        }
    }
    initBranchesPaths(branchesPaths, commit, index, commits) {
        const branch = this.branches.get(commit.branches[0]);
        if (branchesPaths.has(branch)) {
            branchesPaths.set(branch, [
                ...branchesPaths.get(branch),
                { x: commit.x, y: commit.y },
            ]);
        }
        else {
            if (commit.parents[0]) {
                const parentCommit = commits.find(({ hash }) => hash === commit.parents[0]);
                branchesPaths.set(branch, [
                    { x: parentCommit.x, y: parentCommit.y },
                    { x: commit.x, y: commit.y },
                ]);
            }
            else {
                branchesPaths.set(branch, [{ x: commit.x, y: commit.y }]);
            }
        }
        return branchesPaths;
    }
    addMergeCommitsIntoBranchesPaths(commits, branchesPaths) {
        const mergeCommits = commits.filter(({ parents }) => parents.length > 1);
        mergeCommits.forEach(mergeCommit => {
            const parentOnOriginBranch = commits.find(({ hash }) => {
                return hash === mergeCommit.parents[1];
            });
            if (!parentOnOriginBranch)
                return;
            const originBranchName = parentOnOriginBranch.branches[1];
            const branch = this.branches.get(originBranchName);
            if (!branch)
                return;
            const lastPoints = [...(branchesPaths.get(branch) || [])];
            branchesPaths.set(branch, [
                ...lastPoints,
                { x: mergeCommit.x, y: mergeCommit.y, mergeCommit: true },
            ]);
        });
    }
    smoothBranchesPaths(flatBranchesPaths) {
        const branchesPaths = new Map();
        flatBranchesPaths.forEach((points, branch) => {
            if (points.length <= 1) {
                branchesPaths.set(branch, [points]);
                return;
            }
            if (this.isVertical) {
                points = points.sort((a, b) => (a.y > b.y ? -1 : 1));
            }
            else {
                points = points.sort((a, b) => (a.x > b.x ? 1 : -1));
            }
            const paths = points.reduce((mem, point, i) => {
                if (point.mergeCommit) {
                    mem[mem.length - 1].push(pick(point, ["x", "y"]));
                    if (points[i - 1])
                        mem.push([points[i - 1]]);
                }
                else {
                    mem[mem.length - 1].push(point);
                }
                return mem;
            }, [[]]);
            if (this.isVertical) {
                paths.forEach(subPath => {
                    if (subPath.length <= 1)
                        return;
                    const firstPoint = subPath[0];
                    const lastPoint = subPath[subPath.length - 1];
                    const column = subPath[1].x;
                    const branchSize = Math.round(Math.abs(firstPoint.y - lastPoint.y) /
                        this.template.commit.spacing) - 1;
                    const branchPoints = branchSize > 0
                        ? new Array(branchSize).fill(0).map((_, i) => ({
                            x: column,
                            y: subPath[0].y - this.template.commit.spacing * (i + 1),
                        }))
                        : [];
                    const lastSubPaths = branchesPaths.get(branch) || [];
                    branchesPaths.set(branch, [
                        ...lastSubPaths,
                        [firstPoint, ...branchPoints, lastPoint],
                    ]);
                });
            }
            else {
                paths.forEach(subPath => {
                    if (subPath.length <= 1)
                        return;
                    const firstPoint = subPath[0];
                    const lastPoint = subPath[subPath.length - 1];
                    const column = subPath[1].y;
                    const branchSize = Math.round(Math.abs(firstPoint.x - lastPoint.x) /
                        this.template.commit.spacing) - 1;
                    const branchPoints = branchSize > 0
                        ? new Array(branchSize).fill(0).map((_, i) => ({
                            y: column,
                            x: subPath[0].x + this.template.commit.spacing * (i + 1),
                        }))
                        : [];
                    const lastSubPaths = branchesPaths.get(branch) || [];
                    branchesPaths.set(branch, [
                        ...lastSubPaths,
                        [firstPoint, ...branchPoints, lastPoint],
                    ]);
                });
            }
        });
        return branchesPaths;
    }
}
export default GitgraphCore;
//# sourceMappingURL=gitgraph.js.map